import math
import json

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.checkpoint import checkpoint


class OptimusConfig():
    """
    Optimus model configuration class to store the configuration of an
    OptimusTransformer. Left on defaults, will instantiate an OptimusTransformer
    similar in size to GPT2 (~1.5B parameters).

    Args:
        vocab_size (int, defaults to 50257):
            Vocabulary size of the model. Affects the input embeddings and
            output matrix. Make sure to use with a proper tokenizer with the
            correct vocab size.
        hidden_size (int, defaults to 1600):
            Dimension of hidden representations.
        intermediate_size (int, defaults to 4352):
            Dimension of the MLP (feed forward layer) representations. This
            should be equal to 8 * hidden_size / 3, rounded up to a multiple of
            256.
        num_hidden_layers (int, defaults to 48):
            Number of hidden layers in the OptimusTransformer decoder.
        num_attention_heads (int, defaults to 40):
            Number of attention heads for each attentian layer in the
            OptimusTransformer decoder.
        hidden_act (str, defaults to "silu"):
            The non-linear activation function used in the MLP (feed forward)
            layers of the transformer.
        max_position_embeddings (int, defaults to 2048):
            The maximum sequence length that the model might be used for. This
            is used to pre-compute the sinusoidal position embeddings.
        initializer_range (float, defaults to 0.02):
            Standard deviation for initializing weight matrices.
        rms_norm_eps (float, defaults to 1e-6):
            Epsilon value used by the RMS normalization layers.
        tie_word_embeddings (bool, defaults to False):
            Whether to tie word embeddings with the output matrix. Note: This
            has limited support with PyTorch FSDP!
        attn_implementation (str, defaults to "sdpa"):
            Attention implementation. Can be "eager" for the original
            implementation, or "sdpa" for PyTorch's built-in Scaled Dot-Product
            Attention, which implements eager mode, flash attention 2 and
            memory-efficient attention in core.
        attention_bias (bool, defaults to False):
            Whether to use a bias term in the query, key, value and output
            projection matrices in the self-attention layer.
        attention_dropout (float, defaults to 0.0):
            Dropout ratio for the attention probabilities.
        gradient_checkpointing (bool, defaults to True): Whether to enable
            gradient checkpointing (also known as activation checkpointing),
            which trades compute for memory. This checkpoints every transformer
            decoder layer. The memory savings are in the ballpark of
            O(sqrt(n_layers)), while computation increases by ~30% (equivalent
            to running the forward pass twice instead of once).

    """
    def __init__(
        self,
        vocab_size=50257,
        hidden_size=1600,
        intermediate_size=4352,
        num_hidden_layers=48,
        num_attention_heads=40,
        hidden_act='silu',
        max_position_embeddings=2048,
        initializer_range=0.02,
        rms_norm_eps=1e-6,
        tie_word_embeddings=False,
        attn_implementation='sdpa',
        attention_bias=False,
        attention_dropout=0.0,
        gradient_checkpointing=True,
    ):
        self.vocab_size = vocab_size
        self.max_position_embeddings = max_position_embeddings
        self.hidden_size = hidden_size
        self.intermediate_size = intermediate_size
        self.num_hidden_layers = num_hidden_layers
        self.num_attention_heads = num_attention_heads
        self.hidden_act = hidden_act
        self.initializer_range = initializer_range
        self.rms_norm_eps = rms_norm_eps
        self.tie_word_embeddings = tie_word_embeddings
        self.attn_implementation = attn_implementation
        self.attention_bias = attention_bias
        self.attention_dropout = attention_dropout
        self.gradient_checkpointing=gradient_checkpointing
        super().__init__()

    @classmethod
    def from_json_file(cls, file_path):
        with open(file_path, 'r') as file:
            return cls(**json.load(file))


class OptimusRMSNorm(nn.Module):
    """
    Root-mean square layer normalization (https://arxiv.org/abs/1910.07467).

    Args:
        See OptimusTransformer.

    """
    def __init__(self, hidden_size: int, eps: float = 1e-6):
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(hidden_size))

    def _norm(self, x):
        variance = x.pow(2).mean(dim=-1, keepdim=True)
        return x * torch.rsqrt(variance + self.eps)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # compute in float32, not in fp16, since normalization needs to be accurate
        output = self._norm(x.float()).type_as(x)
        return self.weight * output


class OptimusFeedForward(nn.Module):
    """
    Feed forward (MLP) layer, similar to LLama2's implementation of MLP.

    This is a modified version of the original feed forward layer from
    "Attention Is All You Need". The changes are described in "GLU variants
    improve transformer" (https://arxiv.org/abs/2002.05202).

    A summary of the changes proposed:
    - 3 linear layers instead of 2
    - reduce dimension of the 3 matrices to 2/3 * dim to keep computation
      constant
    - change the activation function to SwiGLU (Swish)

    Args:
        See OptimusTransformer.

    """
    def __init__(self, config: OptimusConfig):
        super().__init__()
        hidden_size = config.hidden_size
        intermediate_size = config.intermediate_size

        self.gate_proj = nn.Linear(hidden_size, intermediate_size, bias=False)
        self.up_proj = nn.Linear(hidden_size, intermediate_size, bias=False)
        self.down_proj = nn.Linear(intermediate_size, hidden_size, bias=False)

        if config.hidden_act == 'silu' or config.hidden_act == 'swish':
            self.act_fn = F.silu
        else:
            raise KeyError(f'Currently only silu and swish are supported as activation functions, but got {config.hidden_act}')

    def forward(self, x):
        return self.down_proj(self.act_fn(self.gate_proj(x)) * self.up_proj(x))


class OptimusAttention(nn.Module):
    """
    Multi-head attention layer.

    Args:
        See OptimusTransformer.

    """
    def __init__(self, config: OptimusConfig, layer_idx: int):
        super().__init__()
        self.layer_idx = layer_idx
        self.attention_dropout = config.attention_dropout
        self.hidden_size = config.hidden_size
        self.num_heads = config.num_attention_heads
        self.head_dim = self.hidden_size // self.num_heads

        assert self.hidden_size % self.num_heads == 0
        self.scale = 1 / math.sqrt(self.head_dim)

        self.q_proj = nn.Linear(self.hidden_size, self.num_heads * self.head_dim, bias=config.attention_bias)
        self.k_proj = nn.Linear(self.hidden_size, self.num_heads * self.head_dim, bias=config.attention_bias)
        self.v_proj = nn.Linear(self.hidden_size, self.num_heads * self.head_dim, bias=config.attention_bias)
        self.o_proj = nn.Linear(self.num_heads * self.head_dim, self.hidden_size, bias=config.attention_bias)

    def forward(self,
                x: torch.Tensor,
                mask: torch.Tensor | None = None) -> torch.Tensor:

        bs, seq_len, _ = x.shape # (batch_size, seq_len, dim)

        # linearly project the inputs
        Q = self.q_proj(x) # (batch_size, seq_len, n_heads * head_dim)
        K = self.k_proj(x)
        V = self.v_proj(x)

        # split into num_heads to compute attention
        queries = Q.view(bs, seq_len, self.num_heads, self.head_dim).permute(0, 2, 1, 3) # (batch_size, n_heads, seq_len, head_dim)
        keys = K.view(bs, seq_len, self.num_heads, self.head_dim).permute(0, 2, 1, 3)
        values = V.view(bs, seq_len, self.num_heads, self.head_dim).permute(0, 2, 1, 3)

        # compute attention matmul
        keys = keys.permute(0, 1, 3, 2) # (batch_size, n_heads, head_dim, seq_len)
        scores = queries @ keys # (batch_size, n_heads, seq_len, seq_len)

        # scale
        scores = scores * self.scale

        # mask
        if mask is not None:
            scores = scores + mask

        # softmax (attention probabilities) in float32 + dropout
        attn_probs = F.softmax(scores, dim=-1, dtype=torch.float32).to(dtype=Q.dtype)
        attn_probs = F.dropout(attn_probs, p=self.attention_dropout, training=self.training)

        # matmul
        output = attn_probs @ values # (batch_size, n_heads, seq_len, head_dim)

        output = output.permute(0, 2, 1, 3).contiguous()
        output = output.view(bs, seq_len, self.num_heads * self.head_dim) # (batch_size, seq_len, num_heads * head_dim)

        # final linear
        output = self.o_proj(output)
        return output


class OptimusSdpaAttention(OptimusAttention):
    """
    PyTorch native scaled dot-product attention. If GPU supports it, runs
    Flash Attention 2. Inherits from normal 'OptimusAttention' as the weights
    remain untouched. Only changes the forward pass.

    """
    def forward(self,
                x: torch.Tensor,
                mask: torch.Tensor | None = None) -> torch.Tensor:

        bs, seq_len, _ = x.shape # (batch_size, seq_len, dim)

        # linearly project the inputs
        Q = self.q_proj(x) # (batch_size, seq_len, n_heads * head_dim)
        K = self.k_proj(x)
        V = self.v_proj(x)

        # split into num_heads to compute attention
        queries = Q.view(bs, seq_len, self.num_heads, self.head_dim).permute(0, 2, 1, 3) # (batch_size, n_heads, seq_len, head_dim)
        keys = K.view(bs, seq_len, self.num_heads, self.head_dim).permute(0, 2, 1, 3)
        values = V.view(bs, seq_len, self.num_heads, self.head_dim).permute(0, 2, 1, 3)

        # compute attention through flash attention 2, if supported by the GPU
        attn_output = F.scaled_dot_product_attention(
            queries,
            keys,
            values,
            dropout_p=self.attention_dropout,
            attn_mask=None,
            is_causal=mask is not None,
        )

        output = attn_output.permute(0, 2, 1, 3).contiguous()
        output = output.view(bs, seq_len, self.num_heads * self.head_dim) # (batch_size, seq_len, num_heads * head_dim)

        # final linear
        output = self.o_proj(output)
        return output


OPTIMUS_ATTENTION_CLASSES = {
    'eager': OptimusAttention,
    'sdpa': OptimusSdpaAttention,
}


class OptimusDecoderLayer(nn.Module):
    """
    OptimusTransformer decoder layer; contains a multi-head attention and a feed
    forward (MLP) layer.

    Args:
        See OptimusTransformer.

    """
    def __init__(self, config: OptimusConfig, layer_idx: int):
        super().__init__()
        self.self_attn = OPTIMUS_ATTENTION_CLASSES[config.attn_implementation](config, layer_idx)
        self.ffn = OptimusFeedForward(config)
        self.attn_norm = OptimusRMSNorm(config.hidden_size, config.rms_norm_eps)
        self.ffn_norm = OptimusRMSNorm(config.hidden_size, config.rms_norm_eps)

    def forward(self,
                x: torch.Tensor,
                attn_mask: torch.Tensor,
                ) -> torch.Tensor:
        """
        Compute a layer of the model (attention + ffn).
        """
        # compute normed attention and normed feed forward
        x = x + self.self_attn(self.attn_norm(x), attn_mask)
        x = x + self.ffn(self.ffn_norm(x))
        return x


class OptimusTransformer(nn.Module):
    """
    A plain PyTorch decoder-only GPT-style transformer implementation. Most of
    this model directly copies LLama2's implementation, with small changes and
    adjustments for experimenting and playing around.

    Args:
        config (OptimusConfig): An OptimusTransformer configuration used to
            instantiate the model. With defaults, instantiates a model similar
            to GPT2 XL in size (1.5B).

    """
    def __init__(self, config: OptimusConfig):
        super().__init__()
        self.config = config
        self.vocab_size = config.vocab_size

        self.embeddings = nn.Embedding(config.vocab_size, config.hidden_size)
        self.layers = nn.ModuleList(
            [OptimusDecoderLayer(config, layer_idx) for layer_idx in range(config.num_hidden_layers)]
        )

        self.positional_encodings = self._compute_freqs(config.hidden_size)

        self.output = nn.Linear(config.hidden_size, self.vocab_size, bias=False)
        self.output_norm = OptimusRMSNorm(config.hidden_size, config.rms_norm_eps)

        # ! careful, this has limited FSDP support
        if config.tie_word_embeddings:
            self.output.weight = self.embeddings.weight

        self.apply(self._init_weights)

    def _init_weights(self, module):
        std = self.config.initializer_range
        if isinstance(module, nn.Linear):
            module.weight.data.normal_(mean=0.0, std=std)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=std)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()

    def _compute_freqs(self, dim: int,
                       theta: float = 10000.0) -> torch.Tensor:
        """
        Precompute a frequency matrix to apply to input embeddings (positional
        encoding).

        Args:
            dim (int): Dimension of embeddings in the model.
            theta (float): Frequency.

        Returns:
            torch.Tensor: Positional encodings tensor.

        """
        max_seq_len = self.config.max_position_embeddings
        pos = torch.arange(0, max_seq_len).unsqueeze(1)
        i = torch.arange(0, dim, 2)
        div_term = torch.exp(i * (-math.log(theta) / dim))

        enc = torch.zeros(max_seq_len, dim)
        enc[:, 0::2] = torch.sin(pos * div_term)
        enc[:, 1::2] = torch.cos(pos * div_term)
        enc = enc.unsqueeze(0)

        return enc

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the model to the input tokens.

        Args:
            x (torch.Tensor): Input token indices. Expected in the shape
                (batch_size, seq_len).

        Returns:
            torch.Tensor: Output logits of the model.

        """
        _, seq_len = x.shape # (bs, seq_len)

        # input embeddings and positional encoding
        x = self.embeddings(x) # (batch_size, seq_len, dim)
        encs = self.positional_encodings[:,:seq_len].to(x.device)
        x = x + encs

        # compute mask for masked self-attention
        attn_mask = torch.full((1, seq_len, seq_len), torch.finfo(x.dtype).min)
        attn_mask = torch.triu(attn_mask, diagonal=1).to(dtype=x.dtype, device=x.device)

        # go through all model layers
        for layer in self.layers:
            if self.config.gradient_checkpointing and self.training:
                x = checkpoint(
                    layer.__call__,
                    x,
                    attn_mask,
                    use_reentrant=False,
                ) # (batch_size, seq_len, dim)
            else:
                x = layer(x, attn_mask) # (batch_size, seq_len, dim)

        # final norm + linear layer
        x = self.output_norm(x)
        output = self.output(x) # (batch_size, seq_len, vocab_sz)

        return output
