import os
import logging
from pathlib import Path

import fire
import torch
import torch.distributed as dist
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler

from optimus.trainer import Trainer, TrainingArguments
from optimus.utils import dist_utils, logging_utils, setup_utils


logger = logging.getLogger(__name__)


def main(
    # training args
    batch_size: int = 2, # per GPU
    grad_acc_steps: int = 4, # per GPU
    seq_len: int = 4096,
    lr_max: float = 1e-4,
    weight_decay: float = 0.001,
    warmup_steps: int = 1000,
    epochs: int = 1,
    grad_clip_norm: float = 1.0,
    use_fp16: bool = True,
    seed: int = 42,
    log_on_main_rank_only: bool = True,

    # model + tokenizer
    model_config_path: Path = Path('config.json'),
    hf_tokenizer_name: str = 'gpt2',

    # data directory (as this is usually big, it should reside in a different
    # directory than the source code)
    data_dir: Path = Path('.'),

    # dataset dir (appended to data_dir)
    dataset_dir: Path = Path('gpt2_tokenized_wikitext103_no_seq_len'),

    # training-related dirs (appended to data_dir)
    checkpoints_dir: Path = Path('training_checkpoints'),
    log_dir: Path = Path('training_logs'),
    save_dir: Path = Path('trained_model'),
):
    """
    Prepare the training arguments, then run the training loop.

    Args:
        batch_size (int): Batch size for training. In distributed setup, this is
            per GPU.
        grad_acc_steps (int): Number of batches to accumulate gradients for
            before running backpropagation to update weights. In distributed
            setup, this is per GPU. Global batch size is `batch_size *
            grad_acc_steps * number of GPU's`. Adjust learning rate as needed!
        seq_len (int): Context length for training. This is used to split the
            dataset into batches.
        lr_max (float): Maximum learning rate, used for one-cycle scheduling.
        weight_decay (float): Weight decay used for the optimizer.
        warmup_steps (int): Warmup steps used for the optimizer.
        epochs (int): Number of epochs to train for.
        grad_clip_norm (float): Gradient clipping value for gradient's norm.
        use_fp16 (bool): Whether to train using floating-point 16-bits
            precision. Bfloat16 is used if available, otherwise fp16.
        seed (int): Seed used for reproducibility purposes. Each process will
            have its initial PyTorch seed set to `seed + process_rank`.
        log_on_main_rank_only (bool): Whether output should only be printed by
            the main rank (rank 0).
        model_config_path (Path): Path to a config file which describes the
            model to be trained.
        hf_tokenizer_name (str): HuggingFace tokenizer name.
        data_dir (Path): Directory located in a large/fast storage, which holds
            data to be used by the model. Should also be capable to accomodate
            training checkpoints and logging data. This should not be the same
            as the directory of the source code, as this directory can easily
            reach hundreds of gigabytes on big datasets/model sizes.
        dataset_dir (Path): Path to dataset to be used for training. Note:
            Should be tokenized _with the same tokenizer_ as the one used for
            training (`hf_tokenizer_name` above). The dataset should not be
            already split into batches. Given path is appended to `data_dir`.
        checkpoints_dir (Path): Path where training checkpoints should be saved.
            Will be created if it doesn't exist. Given path is appended to
            `data_dir`.
        log_dir (Path): Path where training logs should be saved. Will be
            created if it doesn't exist. Given path is appended to `data_dir`.
        save_dir (Path): Path where to save the model upon training completion.
            Will be created if it doesn't exist. Given path is appended to
            `data_dir`.

    """
    # set up logging before anything else
    logging.basicConfig(
        level=os.environ.get('LOG_LEVEL', logging.INFO),
        format='{asctime} - {filename} | {levelname} | {message}',
        style='{',
    )

    # create paths
    if isinstance(data_dir, str):
        data_dir = Path(data_dir)

    if isinstance(dataset_dir, str):
        dataset_dir = data_dir / Path(dataset_dir)
    else:
        dataset_dir = data_dir / dataset_dir

    if isinstance(model_config_path, str):
        model_config_path = Path(model_config_path)

    if isinstance(data_dir, str):
        data_dir = Path(data_dir)

    if isinstance(dataset_dir, str):
        dataset_dir = data_dir / Path(dataset_dir)
    else:
        dataset_dir = data_dir / dataset_dir

    if isinstance(checkpoints_dir, str):
        checkpoints_dir = data_dir / Path(checkpoints_dir)
    else:
        checkpoints_dir = data_dir / checkpoints_dir

    if isinstance(log_dir, str):
        log_dir = data_dir / Path(log_dir)
    else:
        log_dir = data_dir / log_dir

    if isinstance(save_dir, str):
        save_dir = data_dir / Path(save_dir)
    else:
        save_dir = data_dir / save_dir

    logger.info(f"Running with:\n"
        f"\t- batch size: {batch_size}\n"
        f"\t- gradient accumulation steps: {grad_acc_steps}\n"
        f"\t- context length: {seq_len}\n"
        f"\t- max learning rate: {lr_max}\n"
        f"\t- weight decay: {weight_decay}\n"
        f"\t- warmup steps: {warmup_steps}\n"
        f"\t- epochs: {epochs}\n"
        f"\t- gradient clipping norm: {grad_clip_norm}\n"
        f"\t- 16-bit floating-point training (fp16): {use_fp16}\n"
        f"\t- seed: {seed}\n"
        f"\t- only main rank logs: {log_on_main_rank_only}\n"
        f"\t- model config file: {model_config_path}\n"
        f"\t- huggingface tokenizer: {hf_tokenizer_name}\n"
        f"\t- training data directory: {str(data_dir)}\n"
        f"\t- dataset directory: {str(dataset_dir)}\n"
        f"\t- checkpoints directory: {str(checkpoints_dir)}\n"
        f"\t- logging directory: {str(log_dir)}\n"
        f"\t- saved model directory: {str(save_dir)}\n"
        f"Please seek '--help' if you want to change any of these settings")

    if log_on_main_rank_only:
        logger.addFilter(logging_utils.FilterMainRankOnly())

    # init distributed process group
    dist_utils.init_process_group()
    torch.manual_seed(seed + dist_utils.get_rank())

    # set device
    device = f'cuda:{dist_utils.get_local_rank()}'

    logger.info(f'Rank {dist_utils.get_rank()}, world size {dist_utils.get_world_size()}, device {device}')

    # set each process's GPU
    torch.cuda.set_device(dist_utils.get_local_rank())
    torch.cuda.empty_cache()

    # load dataset and split into batches
    with dist_utils.main_process_first():
        dataset = setup_utils.load_and_chunk_dataset(str(dataset_dir), seq_len)
        dataset.set_format('torch')

    # load tokenizer
    tokenizer = setup_utils.create_tokenizer(hf_tokenizer_name)

    # create model and move to device
    model = setup_utils.create_model(model_config_path, device)

    # create samplers + dataloaders
    train_dist_sampler = DistributedSampler(
        dataset['train'],
        num_replicas=dist_utils.get_world_size(),
        rank=dist_utils.get_rank(),
        shuffle=True,
        seed=seed,
        drop_last=True,
    )

    train_dataloader = DataLoader(
        dataset['train'],
        batch_size=batch_size, # per GPU
        num_workers=6, # allow pre-fetching data through multi-process
        pin_memory=True, # fast CPU-GPU transfer
        sampler=train_dist_sampler,
    )

    eval_dist_sampler = DistributedSampler(
        dataset['test'],
        num_replicas=dist_utils.get_world_size(),
        rank=dist_utils.get_rank(),
        shuffle=True,
        seed=seed,
        drop_last=True,
    )

    eval_dataloader = DataLoader(
        dataset['test'],
        batch_size=batch_size, # per GPU
        num_workers=6, # allow pre-fetching data through multi-process
        pin_memory=True, # fast CPU-GPU transfer
        sampler=eval_dist_sampler,
    )

    # create optimizer
    optimizer = torch.optim.AdamW(
        model.parameters(),
        betas=(0.9, 0.999),
        eps=1e-7,
        weight_decay=weight_decay,
    )

    # create learning rate scheduler (fastai's implementation)
    lr_scheduler = torch.optim.lr_scheduler.OneCycleLR(
        optimizer=optimizer,
        max_lr=lr_max,
        epochs=epochs,
        steps_per_epoch=len(train_dataloader) // grad_acc_steps,
    )

    # create training arguments
    train_args = TrainingArguments(
        device=torch.device(device),

        # logging
        log_steps=100,
        log_on_main_rank_only=log_on_main_rank_only,

        # core training
        seed=seed,
        optimizer=optimizer,
        lr_scheduler=lr_scheduler,
        num_train_epochs=epochs,
        per_device_batch_size=batch_size, # per GPU
        gradient_accumulation_steps=grad_acc_steps, # per GPU
        max_grad_norm=grad_clip_norm,
        use_fp16=use_fp16,

        # training checkpointing
        checkpoints_dir=checkpoints_dir,
        save_steps=1000,
        save_limit=3,
    )

    # create trainer
    trainer = Trainer(
        train_args=train_args,
        model=model,
        train_dataloader=train_dataloader,
        eval_dataloader=eval_dataloader,
        tokenizer=tokenizer,
    )

    logger.info('Starting training...')

    trainer.train()

    logger.info(f"Finished training! Saving model weights to '{str(save_dir)}'")

    # save model + tokenizer
    trainer.save_model(save_dir)

    # save log data
    trainer.save_logs(log_dir)

    # clean distributed process group
    dist_utils.destroy_process_group()


if __name__=='__main__':
    fire.Fire(main)
